% American or European option call price using PDE solver
% Uses constant implied vol

% Fabrice Douglas Rouah, FRouah.com and Volopta.com

clc; clear;

% Black Scholes call and put
BSC = @(s,K,rf,q,v,T) (s*exp(-q*T)*normcdf((log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T)) - K*exp(-rf*T)*normcdf((log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T) - v*sqrt(T)));
BSP = @(s,K,rf,q,v,T)  K*exp(-rf*T)*normcdf(-(log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T) + v*sqrt(T)) - s*exp(-q*T)*normcdf(-(log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T));

% SPY market maturities, strikes, implied vols
T = [0.123287671 0.268493151 0.715068493 0.953424658];
K = [120 125 130 135 140 145 150];
IV = [...
    0.2780	0.2638	0.2532	0.2518;
    0.2477	0.2402	0.2364	0.2369;
    0.2186	0.2158	0.2203	0.2239;
    0.1878	0.1930	0.2047	0.2098;
    0.1572	0.1712	0.1894	0.1970;
    0.1334	0.1517	0.1748	0.1849;
    0.1323	0.1373	0.1618	0.1736];

% SPY settings
S0 = 137.14;
r = 0.03;
q = 0.0268;

% Choose option flavor and exercise type for the option to price
PutCall = 'P';
EuroAmer = 'A';
Strike = 132.50;
Mat = 7/12;

% Interpolate to get the K/T implied volatility for constant vol model
sigma = interp2(T,K,IV,Mat,Strike);
fprintf('Interpolated volatility is %5.4f \n',sigma);

% Calculate the Black Scholes or Trinomial tree price
if strcmp(EuroAmer,'A')
    N = 250;
    ClosedPrice = Trinomial(S0,Strike,r,q,sigma,Mat,N,PutCall,EuroAmer);
else
    if strcmp(PutCall,'P')
        ClosedPrice = BSP(S0,Strike,r,q,sigma,Mat);
    else
        ClosedPrice = BSC(S0,Strike,r,q,sigma,Mat);
    end
end

%% Grid for the stock price
% Number of stock price steps and maturity steps
nS = 50;
nT = 100;
NS = nS+1;
NT = nT+1;

% Choose the grid type and maturity and stock price limits
GridType = 'NonUniform';
Tmin = 0; 
Tmax = Mat;
Smax = 1.75*Strike;

% Build the stock price grid
if strcmp(GridType,'Uniform')
    Smin = 0.50*Strike;
    ds = (Smax-Smin)/nS;
    S = Smin:ds:Smax;
else
    c = Strike/5;
    dz = 1/nS*(asinh((Smax-Strike)/c) - asinh(-Strike/c));
    for i=1:nS+1;
        z(i) = asinh(-Strike/c) + (i-1)*dz;
        S(i) = Strike + c*sinh(z(i));
    end
end

% Maturity grid -- common to both uniform and non-uniform grid types
dt = (Tmax-Tmin)/nT;
T = [0:NT-1].*dt;


%% Obtain the finite difference prices with the Weighted method

% Explicit (0), Implicit (1), and Crank-Nicolson (0.5) prices
thet = [0,1,0.5];
PDEmethod = {'Explicit   ','Implicit   ','Crank-Nicol'};

% imethod = 'Inverse' slow method uses LU decomposition of A
% imethod = 'Tridiagonal' fast method exploits triangular nature of A
imethod = 'Tridiagonal';
for k=1:length(thet)
    tic
    Price(k) = WeightedMethodBS(thet(k),S0,Strike,sigma,r,q,S,T,dt,GridType,PutCall,EuroAmer,imethod);
    time = toc;
    fprintf('Finished %s in %6.3f seconds \n',PDEmethod{k},time);
end
Error = ClosedPrice - Price;

%% Output everything
fprintf('----------------------------------------------------------\n');
disp([GridType ' grid type'])
fprintf('Stock price grid size of %5.0f \n',NS);
fprintf('Time step grid size of   %5.0f \n',NT);
fprintf('----------------------------------------------------------\n');
fprintf('Method                     Price        Dollar Difference \n');
fprintf('----------------------------------------------------------\n');
if strcmp(EuroAmer,'E')
    fprintf('Black-Scholes        %12.4f \n',ClosedPrice);
else
    fprintf('Trinomial Tree       %12.4f \n',ClosedPrice);
end    
fprintf('Explicit method      %12.4f   %12.4f \n',Price(1),Error(1));
fprintf('Implicit method      %12.4f   %12.4f \n',Price(2),Error(2));
fprintf('Crank-Nicolson       %12.4f   %12.4f \n',Price(3),Error(3));
fprintf('----------------------------------------------------------\n');

